/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "GuitoneDriver.h"

#include "DatabaseDialogManager.h"
#include "WorkspaceDialogManager.h"

#include <QGenericArgument>

GuitoneDriver::GuitoneDriver(int & argc, char** argv)
  : GuitoneCore(argc, argv), dialogManager(0)
{
    connect(
        this, SIGNAL(monotoneBinaryPathSet()),
        this, SLOT(processCommands()),
        Qt::QueuedConnection
    );
}

GuitoneDriver::~GuitoneDriver()
{
    if (dialogManager) delete dialogManager;
}

void GuitoneDriver::processCommands()
{
    QStringList args = arguments();
    // we're not interested in the program's name and "exec"
    args.removeFirst();
    args.removeFirst();

    if (args.size() < 2)
    {
        error("wrong number of arguments");
        return;
    }

    QString path = args.takeFirst();

    if (path == "-")
    {
        dialogManager = new DialogManager(NULL);
    }
    else
    {
        QFileInfo fileInfo(path);
        if (!fileInfo.exists())
        {
            error("non-existant path");
            return;
        }
        path = fileInfo.absoluteFilePath();

        if (fileInfo.isDir())
        {
            dialogManager = new WorkspaceDialogManager(NULL);
            try
            {
                path = MonotoneManager::normalizeWorkspacePath(path);
                APP->manager()->getThreadForWorkspace(path);
            }
            catch (GuitoneException e)
            {
                error(e);
                return;
            }
            qobject_cast<WorkspaceDialogManager *>(dialogManager)->init(path);
        }
        else if (fileInfo.isFile())
        {
            dialogManager = new DatabaseDialogManager(NULL);
            try
            {
                APP->manager()->getThreadForDatabase(path);
            }
            catch (GuitoneException e)
            {
                error(e);
                return;
            }
            qobject_cast<DatabaseDialogManager *>(dialogManager)->init(path);
        }
        else
        {
            error("file is neither a file nor a directory");
            return;
        }
    }

    // if all dialogs have been closed, quit the event loop and exit
    connect(
        dialogManager, SIGNAL(allDialogsClosed()),
        APP, SLOT(quit())
    );

    QString reqMethod = args.takeFirst();

    const QMetaObject * obj = dialogManager->metaObject();
    QMetaMethod method;
    bool found = false;

    for (int i=0; i<obj->methodCount(); i++)
    {
        method = obj->method(i);
        QByteArray sig(method.signature());
        sig = sig.left(sig.indexOf('('));
        if (sig == reqMethod.toLatin1())
        {
            found = true;
            break;
        }
    }

    if (!found)
    {
        error("method signature not found");
        return;
    }

    QList<QByteArray> methodArgs = method.parameterTypes();

    if (methodArgs.size() != args.size())
    {
        error("argument count doesn't match");
        return;
    }

    if (args.size() > 10)
    {
        error("cannot handle more than 10 function arguments");
        return;
    }

    QGenericArgument genArgs[10];
    for (int i=0; i<methodArgs.size(); i++)
    {
        const char * typeName = methodArgs.at(i).constData();
        QVariant::Type type = QVariant::nameToType(typeName);

        // we currently only use / support strings
        switch (type)
        {
            case QVariant::String:      genArgs[i] = Q_ARG(QString, args.at(i)); break;
            case QVariant::Invalid:     I(false);
            default: error(QString("unrecognized type %1").arg(typeName)); return;
        }
    }

    const char * methodName = static_cast<const char *>(reqMethod.toLatin1().data());
    QMetaObject::invokeMethod(dialogManager, methodName, Qt::DirectConnection,
                              genArgs[0], genArgs[1], genArgs[2], genArgs[3],
                              genArgs[4], genArgs[5], genArgs[6], genArgs[7],
                              genArgs[8], genArgs[9]);
}

void GuitoneDriver::error(const QString & msg)
{
    C(msg);
    QMessageBox::critical(NULL, tr("Error"), msg, QMessageBox::Ok, 0, 0);
    quit();
}

