/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "Certs.h"
#include "BasicIOParser.h"

Certs::Certs(QObject * parent, const DatabaseFile & db)
  : QAbstractItemModel(parent), AutomateCommand(0), databaseFile(db)
{}

Certs::~Certs()
{
    certs.clear();
}

void Certs::readCerts(const QString & rev)
{
    certs.clear();
    reset();

    if (rev.size() == 0) return;

    MonotoneTask task(QStringList() << "certs" << rev);
    AutomateCommand::enqueueDatabaseTask(databaseFile, task);
}

void Certs::processTaskResult(const MonotoneTask & task)
{
    if (task.getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task.getDecodedOutput()));
        return;
    }

    BasicIOParser parser(task.getDecodedOutput());
    I(parser.parse());
    StanzaList list = parser.getStanzas();

    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);

        Cert cert;

        bool isItem = false;

        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);

            // we're now only interested in stanzas starting with a "key" entry
            if (j == 0 && entry.sym != "key") break;

            isItem = true;

            if (entry.sym == "key")
            {
                I(entry.vals.size() == 1);
                cert.key = entry.vals.at(0);
                continue;
            }

            if (entry.sym == "name")
            {
                I(entry.vals.size() == 1);
                cert.name = entry.vals.at(0);
                continue;
            }

            if (entry.sym == "value")
            {
                I(entry.vals.size() == 1);
                cert.value = entry.vals.at(0).trimmed();
                continue;
            }

            if (entry.sym == "signature")
            {
                I(entry.vals.size() == 1);
                QString sig = entry.vals.at(0);

                if (sig == "ok")             cert.signature = Cert::Ok;
                else if (sig == "bad")       cert.signature = Cert::Bad;
                else if (sig == "unknown")   cert.signature = Cert::Unknown;
                else
                    W(QString("Unknown cert signature state '%1'.").arg(sig));

                continue;
            }

            if (entry.sym == "trust")
            {
                I(entry.vals.size() == 1);
                QString trust = entry.vals.at(0);

                if (trust == "trusted")          cert.trust = Cert::Trusted;
                else if (trust == "untrusted")   cert.trust = Cert::Untrusted;
                else
                    W(QString("Unknown cert trust '%1'.").arg(trust));

                continue;
            }

            W(QString("Unknown symbol %1.").arg(entry.sym));
        }

        // check if we really processed an item entry
        if (!isItem) continue;
        certs.append(cert);
    }

    // reset any connected view(s)
    reset();

    // signal that we've finished (whoever listens to that)
    emit certsRead(certs);
}

int Certs::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 6;
}

QVariant Certs::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    if (role == Qt::TextAlignmentRole)
    {
        Qt::Alignment flags = Qt::AlignLeft | Qt::AlignTop;
        return QVariant(flags);
    }

    if (role != Qt::DisplayRole)
    {
        return QVariant();
    }

    int row = index.row();
    if (row >= certs.size()) return QVariant();

    Cert cert = certs.at(row);

    switch (index.column())
    {
        case 0: return QVariant(cert.key);
        case 1: return QVariant(cert.name);
        case 2: return QVariant(cert.value);
        case 3:
            switch (cert.signature)
            {
                case Cert::Ok:          return QVariant(tr("ok"));
                case Cert::Bad:         return QVariant(tr("bad"));
                case Cert::Unknown:     return QVariant(tr("unknown"));
            }
        case 4:
            switch (cert.trust)
            {
                case Cert::Trusted:     return QVariant(tr("trusted"));
                case Cert::Untrusted:   return QVariant(tr("untrusted"));
            }
    }

    return QVariant();
}

Qt::ItemFlags Certs::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    // rows shouldn't be selectable or anything else
    return Qt::ItemIsEnabled;
}

QVariant Certs::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Key"));
            case 1: return QVariant(tr("Name"));
            case 2: return QVariant(tr("Value"));
            case 3: return QVariant(tr("Signature"));
            case 4: return QVariant(tr("Trust"));
        }
    }
    return QVariant();
}

int Certs::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return certs.size();
}

QModelIndex Certs::index(int row, int column, const QModelIndex & parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }
    return createIndex(row, column, 0);
}

QModelIndex Certs::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

void Certs::clearData()
{
    certs.clear();
    reset();
}

