/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "InventoryModel.h"

InventoryModel::InventoryModel(QObject * parent)
  : QAbstractItemModel(parent), inventory()
{
    inventory = new Inventory(this);

    connect(
        inventory, SIGNAL(inventoryRead(const QString &)),
        this, SLOT(maybeResetModel(const QString &))
    );

    connect(
        inventory, SIGNAL(invalidWorkspaceFormat(const QString &)),
        this, SIGNAL(invalidWorkspaceFormat(const QString &))
    );

    connect(
        inventory, SIGNAL(beginInsertRows(ModelItem *, int, int)),
        this, SLOT(triggerBeginInsertRows(ModelItem *, int, int))
    );

    connect(
        inventory, SIGNAL(endInsertRows()),
        this, SLOT(triggerEndInsertRows())
    );

    connect(
        inventory, SIGNAL(beginRemoveRows(ModelItem *, int, int)),
        this, SLOT(triggerBeginRemoveRows(ModelItem *, int, int))
    );

    connect(
        inventory, SIGNAL(endRemoveRows()),
        this, SLOT(triggerEndRemoveRows())
    );

    connect(
        inventory, SIGNAL(dataChanged(ModelItem *)),
        this, SLOT(triggerDataChanged(ModelItem *))
    );
}

InventoryModel::~InventoryModel()
{
    I(nodeInsertions.size() == 0);
    delete inventory;
}

void InventoryModel::setWorkspacePath(const WorkspacePath & ws)
{
    workspacePath = ws;
    inventory->setWorkspacePath(ws);
}

void InventoryModel::refresh(const QString & path)
{
    inventory->read(path, false);
}

bool InventoryModel::canFetchMore(const QModelIndex & parent) const
{
    // special handling for the root item
    if (!parent.isValid())
        return true;

    ModelItem * item = static_cast<ModelItem *>
        (parent.internalPointer());
    I(item);

    InventoryItem * invitem = dynamic_cast<InventoryItem *>
        (item);

    // not an inventory item
    if (!invitem)
    {
        D(QString("skipping %1 (not an inventory item)").arg(item->getLabel()));
        return false;
    }

    if (invitem->isExpanded(Inventory::QueryLevel))
    {
        D(QString("skipping %1 (already expanded)").arg(item->getLabel()));
        return false;
    }

    if (invitem->isAboutToBeExpanded())
    {
        D(QString("skipping %1 (about to be expanded)").arg(item->getLabel()));
        return false;
    }

    return true;
}

void InventoryModel::fetchMore(const QModelIndex & parent)
{
    if (!parent.isValid())
        return;

    ModelItem * item = static_cast<ModelItem *>
        (parent.internalPointer());
    I(item);

    InventoryItem * invitem = dynamic_cast<InventoryItem *>
        (item);

    // not an inventory item
    if (!invitem)
    {
        D(QString("skipping %1 (not an inventory item)").arg(item->getLabel()));
        return;
    }

    // again, if this item is already expanded, don't expand it again
    if (invitem->isExpanded(Inventory::QueryLevel) ||
        invitem->isAboutToBeExpanded()) return;

    invitem->setAboutToBeExpanded();

    inventory->read(invitem->getPath());
}

void InventoryModel::maybeResetModel(const QString & queriedPath)
{
    if (queriedPath.isEmpty())
    {
        // we need to call reset() before we emit dataChanged,
        // otherwise the view doesn't know anything of the new
        // index we're going to present her afterwards
        reset();

        // give the root item a proper label, which is in our case the path
        // to the workspace root directory
        QList<ModelItem *> children = inventory->rootItem->getChildren();

        int workspaceRootRow = 0;

        I(children.size() > workspaceRootRow);

        QString displayPath = workspacePath;
        QString homePath = QDir::homePath();
        if (workspacePath.indexOf(homePath) == 0)
        {
            displayPath = QDir(homePath).relativeFilePath(workspacePath);
            displayPath.prepend("~/");
        }
        children[workspaceRootRow]->setLabel(displayPath);

        emit dataChanged(
            index(workspaceRootRow, 0, QModelIndex()),
            index(workspaceRootRow, 1, QModelIndex())
        );
    }
}

QModelIndex InventoryModel::indexFromItem(ModelItem * item, int col) const
{
    if (item->isRoot())
    {
        return QModelIndex();
    }
    return createIndex(item->row(), col, reinterpret_cast<void *>(item));
}

QModelIndex InventoryModel::index(int row, int column, const QModelIndex & parent) const
{
    ModelItem * parentItem;

    if (!parent.isValid())
    {
        parentItem = inventory->rootItem;
    }
    else
    {
        parentItem = static_cast<ModelItem *>(parent.internalPointer());
        I(parentItem);
    }

    ModelItem * childItem = parentItem->child(row);

    if (childItem)
    {
        return createIndex(row, column, childItem);
    }

    return QModelIndex();
}

int InventoryModel::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant InventoryModel::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    I(item);
    return item->data(index.column(), role);
}

Qt::ItemFlags InventoryModel::flags(const QModelIndex & index) const
{
    if (!index.isValid()) return 0;

    QFlags<Qt::ItemFlag> flags = Qt::ItemIsSelectable | Qt::ItemIsEnabled;

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    I(item);

    // no special flags for pseudo items
    PseudoItem * psitem = dynamic_cast<PseudoItem *>(item);
    if (psitem)
    {
        return flags;
    }

    InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);
    if (invitem)
    {
        /*
        // Disabled until we have figured out how to implement renaming properly
        if (item->isNewNode())
        {
            flags |= Qt::ItemIsEditable | Qt::ItemIsDragEnabled;

            if (item->isDirectory())
            {
                flags |= Qt::ItemIsDropEnabled;
            }
        }
        */
    }
    return flags;
}

// the root item needs to store the header for each field
QVariant InventoryModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        return inventory->rootItem->data(section, role);
    }

    return QVariant();
}

QModelIndex InventoryModel::parent(const QModelIndex & index) const
{
    if (!index.isValid())
    {
        return QModelIndex();
    }

    ModelItem * childItem = static_cast<ModelItem *>(index.internalPointer());
    I(childItem);
    ModelItem * parentItem = childItem->parent();

    if (parentItem == inventory->rootItem)
    {
        return QModelIndex();
    }

    I(parentItem);
    return createIndex(parentItem->row(), 0, parentItem);
}

int InventoryModel::rowCount(const QModelIndex & parent) const
{
    ModelItem * parentItem = inventory->rootItem;

    if (parent.isValid())
    {
        parentItem = static_cast<ModelItem *>(parent.internalPointer());
    }

    I(parentItem);
    return parentItem->childCount();
}

void InventoryModel::triggerBeginInsertRows(ModelItem * parent, int start, int end)
{
    beginInsertRows(indexFromItem(parent, 0), start, end);
    nodeInsertions.push(new NodeInsertion(parent, start, end));
}

void InventoryModel::triggerEndInsertRows()
{
    NodeInsertion * insertion = nodeInsertions.pop();

    QStringList paths;
    for (int i=insertion->start; i<=insertion->end; i++)
    {
        ModelItem * child = insertion->parent->child(i);
        InventoryItem * invItem = qobject_cast<InventoryItem *>(child);
        if (!invItem) continue;
        // do not monitor non-existing files
        if (invItem->getFSType() == InventoryItem::None) continue;
        // a little optimization: we don't care about / monitor ignored files
        if (invItem->hasStatus(InventoryItem::Ignored)) continue;
        paths.push_back(invItem->getPath());
    }

    delete insertion;
    emit pathsInserted(paths);

    endInsertRows();
}

void InventoryModel::triggerBeginRemoveRows(ModelItem * parent, int start, int end)
{
    beginRemoveRows(indexFromItem(parent, 0), start, end);

    QStringList paths;
    for (int i=start; i<=end; i++)
    {
        ModelItem * child = parent->child(i);
        InventoryItem * invItem = qobject_cast<InventoryItem *>(child);
        if (!invItem) continue;
        paths.push_back(invItem->getPath());
    }

    emit pathsRemoved(paths);
}

void InventoryModel::triggerEndRemoveRows()
{
    endRemoveRows();
}

void InventoryModel::triggerDataChanged(ModelItem * item)
{
    emit dataChanged(indexFromItem(item, 0), indexFromItem(item, 2));
}

