/***************************************************************************
 *   Copyright (C) 2008 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "InventoryWatcher.h"
#include "InventoryItem.h"
#include "BasicIOParser.h"

InventoryWatcher::InventoryWatcher(QObject * parent)
    : QFileSystemWatcher(parent)
{
    connect(
        this, SIGNAL(directoryChanged(const QString &)),
        this, SLOT(pathChanged(const QString &))
    );

    connect(
        this, SIGNAL(fileChanged(const QString &)),
        this, SLOT(pathChanged(const QString &))
    );
}

InventoryWatcher::~InventoryWatcher() {}

//! TODO: monitor _MTN/options for changes as well
void InventoryWatcher::setWorkspacePath(const WorkspacePath & wp)
{
    workspace = wp;
    QString revPath(workspace + "/_MTN/revision");
    QFile file(revPath);
    if (!file.open(QIODevice::ReadOnly))
    {
        F(QString("cannot open %1/_MTN/revision for reading").arg(workspace));
    }
    QString contents(file.readAll());
    file.close();

    BasicIOParser parser(contents);
    if (!parser.parse())
    {
        F("could not parse contents of _MTN/revision");
    }
    oldRevisionEntries = parser.getStanzas();
    addPath(revPath);
}

void InventoryWatcher::watchPaths(const QStringList & paths)
{
    QStringList watchedPaths = files() + directories();

    foreach (QString path, paths)
    {
        QString fullPath(workspace);
        if (!path.isEmpty()) fullPath += "/" + path;
        if (watchedPaths.indexOf(fullPath) != -1) continue;
        addPath(fullPath);
    }
}

void InventoryWatcher::unwatchPaths(const QStringList & paths)
{
    QStringList watchedPaths = files() + directories();

    foreach (QString path, paths)
    {
        QString fullPath(workspace);
        if (!path.isEmpty()) fullPath += "/" + path;
        if (watchedPaths.indexOf(fullPath) == -1) continue;
        removePath(fullPath);
    }
}

void InventoryWatcher::pathChanged(const QString & path)
{
    I(path.indexOf(workspace) != -1);

    if (path == QString(workspace + "/_MTN/revision"))
    {
        checkForBookkeepChanges();
        // we need to re-add this path since
        // mtn apparently re-creates the file
        addPath(workspace + "/_MTN/revision");
        return;
    }

    QString pathComponent = path.mid(path.indexOf(workspace) + workspace.length() + 1);
    emit changedPath(pathComponent);
}

void InventoryWatcher::checkForBookkeepChanges()
{
    QFile file(workspace + "/_MTN/revision");
    if (!file.open(QIODevice::ReadOnly))
    {
        F(QString("cannot open %1/_MTN/revision for reading").arg(workspace));
    }
    QString contents(file.readAll());
    file.close();

    BasicIOParser parser(contents);
    if (!parser.parse())
    {
        F("could not parse contents of _MTN/revision");
    }
    StanzaList newRevisionEntries = parser.getStanzas();

    QStringList notifiedPaths;

    foreach (Stanza newStanza, newRevisionEntries)
    {
        QString newEntryType = newStanza.at(0).sym;
        // we don't care about patches, since we're notified of them
        // separately; and the other two are uninteresting for us as well
        if (newEntryType == "format_version" ||
            newEntryType == "new_manifest" ||
            newEntryType == "patch")
        {
            continue;
        }

        bool foundOrAlreadyHandled = false;

        for (int i=0; i<oldRevisionEntries.size(); )
        {
            Stanza oldStanza = oldRevisionEntries.at(i);
            QString oldEntryType = oldStanza.at(0).sym;
            if (oldEntryType == "format_version" ||
                oldEntryType == "new_manifest" ||
                oldEntryType == "patch")
            {
                oldRevisionEntries.removeAt(i);
                foundOrAlreadyHandled = true;
                continue;
            }

            if (newEntryType != oldEntryType)
            {
                i++;
                continue;
            }

            if (newEntryType == "old_revision")
            {
                if (oldStanza.at(0).hash != newStanza.at(0).hash)
                {
                    emit changedBaseRevision(newStanza.at(0).hash);
                }
                oldRevisionEntries.removeAt(i);
                foundOrAlreadyHandled = true;
                break;
            }

            // one stanza entries to compare
            if (newEntryType == "delete" ||
                newEntryType == "add_file" ||
                newEntryType == "add_dir")
            {
                if (oldStanza.at(0).vals.at(0) == newStanza.at(0).vals.at(0))
                {
                    oldRevisionEntries.removeAt(i);
                    foundOrAlreadyHandled = true;
                    break;
                }
            }

            // two stanza entries to compare
            if (newEntryType == "rename" || newEntryType == "clear")
            {
                if (oldStanza.at(0).vals.at(0) == newStanza.at(0).vals.at(0) &&
                    oldStanza.at(1).vals.at(0) == newStanza.at(1).vals.at(0))
                {
                    oldRevisionEntries.removeAt(i);
                    foundOrAlreadyHandled = true;
                    break;
                }
            }

            // three stanza entries to compare
            if (newEntryType == "set")
            {
                if (oldStanza.at(0).vals.at(0) == newStanza.at(0).vals.at(0) &&
                    oldStanza.at(1).vals.at(0) == newStanza.at(1).vals.at(0) &&
                    oldStanza.at(2).vals.at(0) == newStanza.at(2).vals.at(0))
                {
                    oldRevisionEntries.removeAt(i);
                    foundOrAlreadyHandled = true;
                    break;
                }
            }

            i++;
        }

        if (foundOrAlreadyHandled) continue;

        QString path = newStanza.at(0).vals.at(0);
        foreach (QString handledPath, notifiedPaths)
        {
            if (path.indexOf(handledPath + "/") != -1)
            {
                foundOrAlreadyHandled = true;
                break;
            }
        }

        if (foundOrAlreadyHandled) continue;

        notifiedPaths.push_back(path);
        emit changedPath(path);
    }

    // now loop through all the left old items;
    // we should only have a subset of items in this list now
    foreach (Stanza oldStanza, oldRevisionEntries)
    {
        QString path = oldStanza.at(0).vals.at(0);

        bool alreadyNotified = false;
        foreach (QString handledPath, notifiedPaths)
        {
            if (path.indexOf(handledPath + "/"))
            {
                alreadyNotified = true;
                break;
            }
        }

        if (alreadyNotified) continue;

        notifiedPaths.push_back(path);
        emit changedPath(path);
    }

    oldRevisionEntries = newRevisionEntries;
}

