/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "MonotoneThread.h"
#include "StdioParser.h"
#include "BasicIOParser.h"

#include <QMetaType>
#include <QTextCodec>

MonotoneTask::MonotoneTask()
{
    init(ByteArrayList(), ByteArrayList());
}

MonotoneTask::MonotoneTask(const MonotoneTask & other)
{
    returnCode = other.returnCode;
    commandNumber = other.commandNumber;
    threadNumber = other.threadNumber;
    finished = other.finished;
    arguments = other.arguments;
    options = other.options;
    output = other.output;
    abortTask = other.abortTask;
    outputEncoding = other.outputEncoding;
}

MonotoneTask::MonotoneTask(const QStringList & args)
{
    init(stringToByteArrayList(args), ByteArrayList());
}

MonotoneTask::MonotoneTask(const QStringList & args, const QStringList & opts)
{
    init(stringToByteArrayList(args), stringToByteArrayList(opts));
}

MonotoneTask::MonotoneTask(const ByteArrayList & args)
{
    init(args, ByteArrayList());
}

MonotoneTask::MonotoneTask(const ByteArrayList & args, const ByteArrayList & opts)
{
    init(args, opts);
}

void MonotoneTask::init(const ByteArrayList & args, const ByteArrayList & opts)
{
    arguments = args;
    options = opts;
    returnCode = -1;
    finished = false;
    abortTask = false;
    outputEncoding = "UTF-8";

    static bool initialized = false;
    if (!initialized)
    {
        qRegisterMetaType<MonotoneTask>("MonotoneTask");
        initialized = true;
    }
}

ByteArrayList MonotoneTask::stringToByteArrayList(const QStringList & list)
{
    ByteArrayList byteArrayList;
    foreach (QString entry, list)
    {
        byteArrayList.append(entry.toUtf8());
    }
    return byteArrayList;
}

QByteArray MonotoneTask::getEncodedInput() const
{
    QByteArray commandLine;

    if (options.size() > 0)
    {
        // currently mtn can only understand key => value option pairs
        I(options.size() % 2 == 0);

        commandLine.append('o');
        for (int i=0, c=options.size(); i<c; i++)
        {
            QByteArray fragment = options.at(i);
            commandLine.append(QString::number(fragment.size()).toAscii());
            commandLine.append(':');
            commandLine.append(fragment);
        }
        // separate options from the command by a single whitespace
        commandLine.append("e ");
    }

    I(arguments.size() > 0);

    commandLine.append('l');
    for (int i=0, c=arguments.size(); i<c; i++)
    {
        QByteArray fragment = arguments.at(i);
        commandLine.append(QString::number(fragment.size()).toAscii());
        commandLine.append(':');
        commandLine.append(fragment);
    }
    commandLine.append('e');

    return commandLine;
}

QString MonotoneTask::getDecodedOutput(const QString & enc) const
{
    QString encoding = enc;
    if (encoding.isEmpty())
    {
        encoding = outputEncoding;
    }
    I(!encoding.isEmpty());

    // QTextCodec implements codecForName as factory and caches multiple
    // requests, so we don't need to do that ourselves
    QTextCodec * codec = QTextCodec::codecForName(encoding.toLatin1());
    I(codec);
    return codec->toUnicode(output);
}


const int MonotoneThread::StdioBufferSize = 50 * 1024 * 1024;

MonotoneThread::MonotoneThread(
    int thread, const QString & m, const QString & d, const QString & w
) : QThread(), doAbort(false), commandNumber(0), threadNumber(thread),
    mtnBinary(m), databasePath(d), workspacePath(w)
{}

MonotoneThread::~MonotoneThread() {}

int MonotoneThread::enqueueTask(const MonotoneTask & task)
{
    QMutexLocker locker(&lock);

    if (doAbort) return -1;

    commandNumber++;
    MonotoneTask workItem(task);
    workItem.setCommandNumber(commandNumber);
    workItem.setThreadNumber(threadNumber);
    queue.enqueue(workItem);

    waitForTasks.wakeAll();
    D(QString("thread %1: new task enqueued: %2").arg(threadNumber).arg(QString(task.getEncodedInput())));
    return commandNumber;
}

void MonotoneThread::removeTasks(QList<int> commandNumbers)
{
    QMutexLocker locker(&lock);

    for (int i=0; i<queue.size(); i++)
    {
        if (commandNumbers.contains(queue[i].getCommandNumber()))
        {
            queue[i].setAborted();
        }
    }
}

void MonotoneThread::run()
{
    // use a local object instance to automatically trigger the
    // destruction of the process if we leave run()
    QObject threadParent;
    QProcess * process = new QProcess(&threadParent);

    QStringList args;
    args << "automate";
    args << "stdio";
    args << QString("--automate-stdio-size=%1").arg(StdioBufferSize);
    args << "--db" << databasePath;

    // check whether we need to work on a specific workspace directory or not
    if (!workspacePath.isEmpty())
    {
        process->setWorkingDirectory(workspacePath);
    }
    else
    {
        // we need to ensure that the current directory is no workspace
        // directory, so paths won't get interpreted wrongly
        QString tmpPath = QDir::tempPath();
        if (QDir(tmpPath + "/_MTN").exists())
        {
            // this is actually not a QProcess error, but who cares...
            emit aborted(threadNumber, QProcess::UnknownError,
                tr("The temporary directory '%1' contains a monotone "
                   "workspace, which may lead to serious path resolution "
                   "errors on execution. Its recommended you remove "
                   "this workspace before you retry this again!").arg(tmpPath));

            cleanup(process);
            return;
        }
        process->setWorkingDirectory(tmpPath);
    }

    // remove any LANG setting so we get mtn's error output in native English
    // (this is hacky, of course, but there is still no error code
    // implementation in mtn, so we have to do raw string parsing...)
    QStringList env = QProcess::systemEnvironment();
    env.replaceInStrings(QRegExp("^LANG=.*", Qt::CaseInsensitive), "");
    env.push_back("LANG=c");
    process->setEnvironment(env);

    L(QString("starting %1 %2").arg(mtnBinary).arg(args.join(" ")));
    process->start(mtnBinary, args);

    if (!process->waitForStarted())
    {
        emit aborted(threadNumber, process->error(),
            QString::fromUtf8(process->readAllStandardError())
        );
        cleanup(process);
        return;
    }

    // waitForReadyRead returns false, if there are no data available
    // (stdio shouldn't return anything on either channel if everything is ok)
    // or the time is up
    // FIXME: find the proper timeout for our use case here or better an
    // alternative way to detect if something went wrong while setting up
    // the process _before_ we send the whole thread to sleep later on
    // waiting silently for his first task...
    process->setReadChannel(QProcess::StandardError);
    if (process->waitForReadyRead(1000))
    {
        QString err = QString::fromUtf8(process->readAllStandardError());

        emit aborted(threadNumber, process->error(), err);
        cleanup(process);
        return;
    }
    process->setReadChannel(QProcess::StandardOutput);

    D(QString("thread %1: process started").arg(threadNumber));
    emit started(threadNumber);

    QByteArray buffer;
    QByteArray output;

    bool processingTask = false;

    while (!doAbort)
    {
        // send the thread to sleep if there are no tasks to process
        lock.lock();
        if (queue.size() == 0)
            waitForTasks.wait(&lock);
        lock.unlock();

        // check if we got notified to abort the processing
        // while we were sleeping...
        if (doAbort)
            break;

        if (process->state() != QProcess::Running)
        {
            emit aborted(threadNumber, process->error(),
                         QString::fromUtf8(process->readAllStandardError())
            );
            cleanup(process);
            return;
        }

        MonotoneTask task = queue.head();

        // check if we should cancel the processing of this task
        if (task.isAborted())
        {
            processingTask = false;
            buffer.clear();
            output.clear();
            queue.dequeue();

            emit taskAborted(task);
            continue;
        }

        if (!processingTask)
        {
            QByteArray in(task.getEncodedInput());

            if (process->write(in) == -1)
            {
                QByteArray err;
                err.append("could not write to process (stderr: ");
                err.append(process->readAllStandardError());
                err.append(")");

                task = queue.dequeue();
                task.setOutput(err);
                task.setReturnCode(-1);

                emit taskAborted(task);
                continue;
            }

            buffer.clear();
            output.clear();
            processingTask = true;

            D(QString("thread %1: task %2 (%3) started").arg(threadNumber)
              .arg(task.getCommandNumber()).arg(QString(task.getEncodedInput())));
        }

        // FIXME: make this configurable!
        if (!process->waitForReadyRead(30000))
        {
            emit aborted(threadNumber, process->error(),
                         QString::fromUtf8(process->readAllStandardError())
            );
            cleanup(process);
            return;
        }

        // we can't assume that all stderr output from stdio is evil
        // just because there are some automate commands which are
        // programmed badly (f.e. genkey, which outputs status messages
        // on stderr)
        QByteArray err = process->readAllStandardError();
        if (err.size() != 0)
        {
            // FIXME: expand this list further or wait for nvm.automate_out_of_band to land
            QStringList cmdsWithSpuriousOutput =
                QStringList() << "genkey" << "select" << "inventory";
            QString cmd = task.getArguments().at(0);

            if (!cmdsWithSpuriousOutput.contains(cmd))
            {
                task.setOutput(err);
                task.setReturnCode(-1);

                processingTask = false;
                buffer.clear();
                output.clear();
                queue.dequeue();

                emit taskAborted(task);
                continue;
            }
        }

        buffer.append(process->readAllStandardOutput());
        StdioParser parser(buffer);

        // if the chunk is not yet complete, try again later
        if (!parser.parse())
        {
            L("cannot parse - continueing");
            continue;
        }

        buffer = parser.getLeftBytes();
        output.append(parser.getPayload());
        int returnCode = parser.getErrorCode();

        // TODO: support for other chunk types here?
        if (parser.getChunkType() == 'm')
        {
            continue;
        }

        task.setOutput(output);
        task.setReturnCode(returnCode);
        task.setFinished();

        queue.dequeue();

        processingTask = false;

        D(QString("thread %1: task %2 (%3) finished").arg(threadNumber)
            .arg(task.getCommandNumber()).arg(QString(task.getEncodedInput())));

        emit taskFinished(task);
    }
    cleanup(process);
}

void MonotoneThread::cleanup(QProcess * proc)
{
    QMutexLocker locker(&lock);

    doAbort = true;

    if (queue.size() > 0)
    {
        foreach (MonotoneTask task, queue)
        {
            emit taskAborted(task);
        }
        queue.clear();
    }

    // close the pipes
    proc->close();
    // send SIGTERM
    proc->terminate();
    // block until the process has really been finished
    proc->waitForFinished();
}

void MonotoneThread::abort()
{
    QMutexLocker locker(&lock);
    doAbort = true;
    waitForTasks.wakeAll();
}

