/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "MonotoneUtil.h"
#include "GuitoneCore.h"
#include "BasicIOParser.h"

MonotoneUtil::MonotoneUtil(MonotoneThread * thread, const MonotoneTask & in)
{
    connect(
        thread, SIGNAL(taskFinished(const MonotoneTask &)),
        this, SLOT(taskFinished(const MonotoneTask &))
    );
    commandNumber = thread->enqueueTask(in);
}

MonotoneTask & MonotoneUtil::getTaskResult()
{
    eventLoop.exec();
    return result;
}

void MonotoneUtil::taskFinished(const MonotoneTask & task)
{
    if (task.getCommandNumber() != commandNumber) return;
    result = task;
    eventLoop.quit();
}

void MonotoneUtil::taskAborted(const MonotoneTask & task)
{
    if (task.getCommandNumber() != commandNumber) return;
    C(QString("Task %1 (thread %2) aborted (%3)")
        .arg(task.getCommandNumber())
        .arg(task.getThreadNumber())
        .arg(QString::fromUtf8(task.getEncodedInput()))
    );
    result = task;
    eventLoop.quit();
}

//
// static methods
//

MonotoneTask MonotoneUtil::runSynchronousWorkspaceTask(const WorkspacePath & workspace, const MonotoneTask & task)
{
    MonotoneThread * thread = APP->manager()->getThreadForWorkspace(workspace);
    MonotoneUtil * instance = new MonotoneUtil(thread, task);
    MonotoneTask result = instance->getTaskResult();
    delete instance;
    return result;
}

MonotoneTask MonotoneUtil::runSynchronousDatabaseTask(const DatabaseFile & database, const MonotoneTask & task)
{
    MonotoneThread * thread = APP->manager()->getThreadForDatabase(database);
    MonotoneUtil * instance = new MonotoneUtil(thread, task);
    MonotoneTask result = instance->getTaskResult();
    delete instance;
    return result;
}

DatabaseFile MonotoneUtil::getDatabaseFile(const WorkspacePath & workspace)
{
    return APP->manager()->getDatabaseFilePath(workspace);
}

QString MonotoneUtil::getBaseWorkspaceRevision(const WorkspacePath & workspace)
{
    MonotoneTask in(QStringList() << "get_base_revision_id");
    MonotoneTask out = runSynchronousWorkspaceTask(workspace, in);
    if (!out.isFinished()) F("task aborted");

    QString data = out.getDecodedOutput();
    if (out.getReturnCode() > 0)
    {
        C(QString("Could not execute get_base_revision_id: %1").arg(data));
        return QString();
    }

    data.chop(1);

    // we cannot handle workspaces with multiple parents yet
    I(data.indexOf('\n') == -1);

    return data;
}

QString MonotoneUtil::getOption(const WorkspacePath & workspace, const QString & opt)
{
    MonotoneTask in(QStringList() << "get_option" << opt);
    MonotoneTask out = runSynchronousWorkspaceTask(workspace, in);
    if (!out.isFinished()) F("task aborted");

    QString data = out.getDecodedOutput();
    if (out.getReturnCode() > 0)
    {
        C(QString("Couldn't retrieve option %1: %2").arg(opt).arg(data));
        return QString();
    }

    // remove the trailing newline
    data.chop(1);

    return data;
}

QString MonotoneUtil::getBranchName(const WorkspacePath & workspace, const QString & defaultBranch)
{
    QString branchName = getOption(workspace, "branch");

    if (branchName.size() == 0)
    {
        branchName = defaultBranch;
    }
    return branchName;
}

QString MonotoneUtil::getBranchNameShort(const WorkspacePath & workspace)
{
    // FIXME: currently there is only support for the domain branch scheme
    // tld.domain.project.branch
    QString branchName = getBranchName(workspace);
    QStringList parts = branchName.split('.');
    if (parts.size() == 1) return branchName;

    int maxSize = 30;
    int minNeededChars = (parts.size() * 2) - 1;

    QString shortBranchName;
    for (int i=parts.size()-1; i>=0 ; i--)
    {
        QString part = parts.at(i);
        if ((part.size() + shortBranchName.size() + minNeededChars) > maxSize)
        {
            shortBranchName.prepend(part.at(0));
        }
        else
        {
            shortBranchName.prepend(part);
        }

        if (i != 0) shortBranchName.prepend(".");
        minNeededChars -= 2;
    }

    return shortBranchName;
}

QStringList MonotoneUtil::resolveSelector(const DatabaseFile & db, const QString & selector)
{
    MonotoneTask in(QStringList() << "select" << selector);
    MonotoneTask out = runSynchronousDatabaseTask(db, in);
    if (!out.isFinished()) F("task aborted");

    QString data = out.getDecodedOutput();
    if (out.getReturnCode() > 0)
    {
        C(QString("Couldn't resolve selector %1: %2").arg(selector).arg(data));
        return QStringList();
    }

    QStringList revList = data.split('\n', QString::SkipEmptyParts);
    return revList;
}

RevisionCerts MonotoneUtil::getRevisionCerts(const DatabaseFile & db, const QString & revision)
{
    MonotoneTask in(QStringList() << "certs" << revision);
    MonotoneTask out = runSynchronousDatabaseTask(db, in);
    if (!out.isFinished()) F("task aborted");

    QString data = out.getDecodedOutput();
    if (out.getReturnCode() > 0)
    {
        C(QString("Couldn't query revision certs for %1: %2").arg(revision).arg(data));
        return RevisionCerts();
    }

    BasicIOParser parser(data);
    if (!parser.parse())
    {
        C("Could not parse basic_io.");
        return RevisionCerts();
    }

    RevisionCerts certs;
    StanzaList stanzas = parser.getStanzas();

    foreach (Stanza st, stanzas)
    {
        RevisionCert cert;

        foreach (StanzaEntry en, st)
        {
            if (en.sym == "name")
            {
                cert.first = en.vals.at(0);
            }

            if (en.sym == "value")
            {
                cert.second = en.vals.at(0);
            }
        }
        I(!cert.first.isEmpty() && !cert.second.isEmpty());
        certs.append(cert);
    }
    return certs;
}

FileEntryList MonotoneUtil::getRevisionManifest(const DatabaseFile & db, const QString & revision)
{
    MonotoneTask in(QStringList() << "get_manifest_of" << revision);
    MonotoneTask out = runSynchronousDatabaseTask(db, in);
    if (!out.isFinished()) F("task aborted");

    QString data = out.getDecodedOutput();
    if (out.getReturnCode() > 0)
    {
        C(QString("Couldn't query manifest entries for %1: %2").arg(revision).arg(data));
        return FileEntryList();
    }

    BasicIOParser parser(data);
    if (!parser.parse())
    {
        C("Could not parse basic_io.");
        return FileEntryList();
    }

    // FIXME: this should go into a separate function / class
    FileEntryList entries;
    StanzaList stanzas = parser.getStanzas();
    foreach (Stanza st, stanzas)
    {
        FileEntry entry;
        bool is_item = false;

        foreach (StanzaEntry en, st)
        {
            if (en.sym == "format_version")
            {
                I(en.vals.size() == 1 && en.vals.at(0) == "1");
                break;
            }
            else
            if (en.sym == "dir" || en.sym == "file")
            {
                is_item = true;
                entry.path = en.vals.at(0);
                entry.is_dir = en.sym == "dir";
            }
            else
            if (en.sym == "content")
            {
                I(is_item);
                entry.fileid = en.hash;
            }
            else
            if (en.sym == "attr")
            {
                I(is_item);
                I(en.vals.size() == 2);
                entry.attrs.insert(en.vals.at(0), en.vals.at(1));
            }
            else
                I(false);
        }
        if (is_item) entries.append(entry);
    }
    return entries;
}

QStringList MonotoneUtil::getPrivateKeyList(const DatabaseFile & db)
{
    MonotoneTask in(QStringList() << "keys");
    MonotoneTask out = runSynchronousDatabaseTask(db, in);
    if (!out.isFinished()) F("task aborted");

    QString data = out.getDecodedOutput();
    if (out.getReturnCode() > 0)
    {
        C(QString("Couldn't query keys: %1").arg(data));
        return QStringList();
    }

    BasicIOParser parser(data);
    if (!parser.parse())
    {
        C("Could not parse basic_io.");
        return QStringList();
    }

    QStringList keys;
    StanzaList stanzas = parser.getStanzas();
    foreach (Stanza st, stanzas)
    {
        QString lastKey;
        foreach (StanzaEntry en, st)
        {
            if (en.sym == "name")
            {
                lastKey = en.vals.at(0);
                continue;
            }

            if (en.sym == "private_location")
            {
                I(!lastKey.isEmpty());
                keys.append(lastKey);
                break;
            }
        }
    }
    return keys;
}

QString MonotoneUtil::getFileId(const DatabaseFile & db, const QString & path)
{
    MonotoneTask in(QStringList() << "identify" << path);
    MonotoneTask out = runSynchronousDatabaseTask(db, in);
    if (!out.isFinished()) F("task aborted");

    QString data = out.getDecodedOutput();
    if (out.getReturnCode() > 0)
    {
        C(QString("Couldn't identify path: %1").arg(data));
        return QString();
    }

    data.chop(1);
    return data;
}

bool MonotoneUtil::getAttribute(const WorkspacePath & ws, const QString & path,
                  const QString & attrname, QPair<QString, QString> & attrval)
{
    MonotoneTask in(QStringList() << "get_attributes" << path);
    MonotoneTask out = runSynchronousWorkspaceTask(ws, in);
    if (!out.isFinished()) F("task aborted");

    QString data = out.getDecodedOutput();
    if (out.getReturnCode() > 0)
    {
        C(QString("Couldn't run get_attributes: %1").arg(data));
        return false;
    }

    BasicIOParser parser(data);
    I(parser.parse());
    StanzaList stanzas = parser.getStanzas();

    foreach (Stanza st, stanzas)
    {
        if (st.at(0).sym != "attr") continue;
        if (st.at(0).vals.at(0) == attrname)
        {
            attrval.first = st.at(0).vals.at(1);
            attrval.second = st.at(1).vals.at(0);
            return true;
        }
    }

    return false;
}

bool MonotoneUtil::getAttribute(const DatabaseFile & db, const QString & revision,
                  const QString & path, const QString & attrname, QString & attrval)
{
    FileEntryList entryList = getRevisionManifest(db, revision);

    foreach (FileEntry en, entryList)
    {
        if (en.path != path) continue;
        if (en.attrs.contains(attrname))
        {
            attrval = en.attrs.value(attrname);
            return true;
        }
    }

    return false;
}

/*!
    Strips mtn: warning: and alike from error strings coming from mtn
    and unwraps them, if this fails we add the original string unwrapped
*/
QString MonotoneUtil::stripMtnPrefix(const QString & input)
{
    QString output;
    QStringList list = input.split(QRegExp("\\n"));

    // FIXME: we should actually use basename(mtnBinaryPath) in case
    // the binary is renamed, but I'm too lazy for this right now...
    // and this would also require to make this method non-static
    QRegExp rx = QRegExp("^(?:mtn\\:[ ])?\\w+\\:[ ]+(.+)");
    rx.setPatternSyntax(QRegExp::RegExp2);
    for (int i=0, j=list.size(); i<j; i++)
    {
        if (rx.indexIn(list[i]) != -1)
        {
            output += rx.cap(1) + " ";
            continue;
        }
        output += list[i] + " ";
    }
    return output.trimmed();
}

