/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "DiffParser.h"
#include "vocab.h"

#include <QStringList>
#include <QRegExp>

DiffParser::DiffParser(const QString & input)
{
    parse(input);
}

DiffParser::~DiffParser()
{
    qDeleteAll(fileDiffs);
    fileDiffs.clear();
}

void DiffParser::parse(const QString & input)
{
    if (input.length() == 0)
    {
        W("Nothing to parse");
        return;
    }

    // remove last newline character
    QString in(input);
    in.chop(1);

    // split the output into lines
    QStringList lines = in.split(QChar('\n'));

    QString curFile;

    Diff * curDiff = 0;
    DiffHunk * curHunk = 0;

    for (int i=0, s=lines.size(); i<s; i++)
    {
        QString line(lines.at(i));
        I(line.length() > 0);
        QChar curChar = line.at(0);

        //
        // parse file diff separators, setup new DiffFile object
        //
        if (curChar == '=')
        {
            I(i < (s-1));

            // straightly go to the next line
            QString nextGroup(lines.at(++i));

            // create a new diff
            curDiff = new Diff();

            // check if this is a binary
            QRegExp rx = QRegExp("^#\\s(.+)\\sis binary");
            if (rx.indexIn(nextGroup) != -1)
            {
                curDiff->is_binary = true;
                curFile = rx.cap(1);
            }
            else
            {
                // then this should be a normal file diff
                rx = QRegExp("^---\\s([^\\t]+)\\t\\w{40}");
                I(rx.indexIn(nextGroup) > -1);
                curFile = rx.cap(1);
                curDiff->is_binary = false;
                // skip the next line as we don't care about renames
                // yet and if we care about them, we get this info from
                // the cset
                i++;
            }

            fileDiffs.insert(curFile, curDiff);
            continue;
        }

        I(curDiff);

        //
        // parse hunk lines, setup new hunk object
        //
        if (curChar == '@')
        {
            QRegExp rx = QRegExp("^@@\\s+\\-(\\d+)(,(\\d+))?\\s+\\+(\\d+)(,(\\d+))?\\s+@@");
            I(rx.indexIn(line) > -1);
            QStringList list = rx.capturedTexts();
            I(list.size() >= 2);

            // create a new Hunk
            curHunk = new DiffHunk();
            curDiff->hunks.push_back(curHunk);

            // extract line numbers and positions
            curHunk->leftStart = list.at(1).toInt();
            if (list.at(2).length() > 0)
                curHunk->leftCount = list.at(3).toInt();
            else
                curHunk->leftCount = 1;

            curHunk->rightStart = list.at(4).toInt();
            if (list.at(5).length() > 0)
                curHunk->rightCount = list.at(6).toInt();
            else
                curHunk->rightCount = 1;

            continue;
        }

        I(curHunk);

        DiffLine::State state;

        if (curChar == '+') state = DiffLine::Added;
        else if (curChar == '-') state = DiffLine::Removed;
        else if (curChar == ' ') state = DiffLine::Unchanged;
        else if (curChar == '\\')
        {
            // if we have a "no newline at the end of the file" line we
            // also always have an ancestor line which has missing this
            // particular \n characterkate 
            I(curHunk->lines.size() > 0);

            DiffLine * lastLine = curHunk->lines.last();
            switch (lastLine->state)
            {
                case DiffLine::Added:
                    lastLine->state = DiffLine::AddedMissingNewline;
                    break;
                case DiffLine::Removed:
                    lastLine->state = DiffLine::RemovedMissingNewline;
                    break;
                case DiffLine::Unchanged:
                    lastLine->state = DiffLine::UnchangedMissingNewline;
                    break;
                default: I(false);
            }

            // we do not create a new diff line for this pseudo line
            continue;
        }
        else I(false);

        DiffLine * diffLine = new DiffLine();
        curHunk->lines.push_back(diffLine);
        diffLine->content = line.mid(1);
        diffLine->state = state;
    }
}

Diff * DiffParser::getDiff(const QString & filename)
{
    if (!fileDiffs.contains(filename)) return 0;
    return fileDiffs.value(filename);
}

