/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "vocab.h"
#include "Platform.h"
#include "Settings.h"

#include <QApplication>
#include <QFileInfo>
#include <QMessageBox>
#include <QProcess>

#ifdef Q_WS_WIN
#include <windows.h>
#include <lmcons.h>
#endif

#ifdef Q_WS_X11
#include <cstdlib>
using namespace std;
#endif

//
// The following code is public domain, with slight modifications, taken from
// http://hrabia.ibib.waw.pl/~winnie/qtopen.html
// It opens a file with the associated program on the specific platform
// Original author: Chris Thompson <chris@hypocrite.org>
//
bool Platform::openFile(const QString & filename, bool isSymlink /* = false */)
{
    // if this looks like a file, and not like an URL or something else
    if (filename.indexOf(QRegExp("^\\w+:\\/\\/")) == -1)
    {
        QFileInfo fileInfo(filename);
        if (!fileInfo.exists())
        {
            QMessageBox::critical(
                NULL,
                QObject::tr("Error"),
                QObject::tr("The file you're trying to open does not exist."),
                QMessageBox::Ok, 0, 0
            );
            return false;
        }

        if (fileInfo.isSymLink())
        {
            if (isSymlink)
            {
                QMessageBox::critical(
                    NULL,
                    QObject::tr("Error"),
                    QObject::tr("No support for nested symlinks - "
                                "abort opening of %1").arg(filename),
                    QMessageBox::Ok, 0, 0
                );
                return false;
            }
            return openFile(fileInfo.symLinkTarget(), true);
        }

        if (Settings::getBool("AskFileOpen", true))
        {
            QString suffix = fileInfo.completeSuffix();
            QStringList harmfulSuffixes =
                Settings::getString("AskFileOpenExt", ASK_OPEN_FILE_EXT)
                    .split(",", QString::SkipEmptyParts);

            if (fileInfo.isExecutable() || harmfulSuffixes.contains(suffix))
            {
                QMessageBox::StandardButton btn = QMessageBox::question(
                    NULL,
                    QObject::tr("Attention"),
                    QObject::tr("The file '%1' you're trying to open is either "
                                "executable or included in the list of "
                                "precarious file formats - do you really "
                                "want to continue?").arg(filename),
                    QMessageBox::Yes | QMessageBox::No
                );

                if (btn == QMessageBox::No)
                {
                    return false;
                }
            }
        }
    }

    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    int retval = -1;

#ifdef Q_WS_X11
    // due to a bug in KDE < 3.5.3 we can't trust the return value kfmclient
    // gives us back (its always 1!)
    if (system("which kfmclient >/dev/null 2>&1") == 0)
    {
        D("Using kfmclient");
        string s("kfmclient exec \"");
        s += filename.toUtf8().data();
        s += "\"";
        // just cross your fingers
        system(s.c_str());
        retval = 0;
    }
    else
    if (system("which gnome-open >/dev/null 2>&1") == 0)
    {
        D("Using gnome-open");
        string s("gnome-open \"");
        s += filename.toUtf8().data();
        s += "\"";
        retval = system(s.c_str());
    }
    else
    {
        C("Neither kfmclient nor gnome-open found");
    }
#endif

#ifdef Q_WS_MACX
    retval = QProcess::execute("open", QStringList() << filename);
    // if a regular open failed, try to open the file in the system's
    // default editor
    if (retval != 0)
    {
        D(QString("try to open %1 in text editor").arg(filename));
        retval = QProcess::execute("open", QStringList() << "-t" << filename);
    }
#endif

#ifdef Q_WS_WIN
    QString winfilename(filename);
    winfilename.replace(QChar('/'), QString("\\"));
    int code = reinterpret_cast<int>(
        ShellExecute(NULL, L"open", winfilename.toStdWString().data(),
                     NULL, NULL, SW_SHOW)
    );

    // if shellexec fails (return value < 32, see
    // http://msdn2.microsoft.com/en-us/library/bb762153.aspx)
    // f.e. because the file's extension is not registered,
    // try to open in with the "Open With" dialog
    if (code > 32)
    {
        retval = 0;
    }
    else
    {
        QStringList arguments;
        arguments << "shell32.dll,OpenAs_RunDLL" << winfilename;
        retval = QProcess::execute("rundll32.exe", arguments);
    }
#endif

    QApplication::restoreOverrideCursor();

    if (retval != 0)
    {
        QMessageBox::critical(
            NULL,
            QObject::tr("Error"),
            QObject::tr("Unable to open files on your platform - please "
                        "contact the author about this problem."),
            QMessageBox::Ok, 0, 0
        );
        return false;
    }

    return true;
}

//
// the following code is loosely based on David Wright's code snippet
// on the Qt interest mailing list
// (http://lists.trolltech.com/qt-interest/2002-12/thread00130-0.html)
//
QString Platform::getUsername()
{
    QString username;

#ifdef Q_WS_WIN
#if defined(UNICODE)
    if (QSysInfo::WindowsVersion & QSysInfo::WV_NT_based)
    {
        TCHAR winUserName[UNLEN + 1]; // UNLEN is defined in LMCONS.H
        DWORD winUserNameSize = sizeof(winUserName);
        GetUserNameW(winUserName, &winUserNameSize);
        username = QString::fromStdWString(winUserName);
    } else
#endif
    {
        char winUserName[UNLEN + 1]; // UNLEN is defined in LMCONS.H
        DWORD winUserNameSize = sizeof(winUserName);
        GetUserNameA(winUserName, &winUserNameSize);
        username = QString::fromLocal8Bit(winUserName);
    }
#endif

#ifdef Q_WS_X11
#define IS_UNIX
#endif

#ifdef Q_WS_MACX
#define IS_UNIX
#endif

#ifdef IS_UNIX
    username = QString::fromAscii(getenv("USER"));
#endif

    D(QString("Found username %1").arg(username));

    return username;
}

