/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "CheckoutRevision.h"
#include "MonotoneUtil.h"
#include "Settings.h"
#include "WorkspaceCreator.h"

#include <QMessageBox>

CheckoutRevision::CheckoutRevision(QWidget * parent, const DatabaseFile & database)
  : Dialog(parent), databaseFile(database)
{
    setupUi(this);
    Dialog::init();

    branchesModel = new Branches(this, databaseFile, false);
    branchesModel->readBranches();
    branchList->setModel(branchesModel);

    tagsModel = new Tags(this, databaseFile);
    tagsModel->readTags(QString());
    tagList->setModel(tagsModel);

    selectorCompleter = new QCompleter(
         Settings::getItemList("RecentCheckoutSelectors"),
         this
    );

    selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
    selectedRevision->setCompleter(selectorCompleter);

    localCheckoutPath->setText(QDir::homePath());

    connect(
        selectRevision, SIGNAL(clicked()),
        this, SLOT(slotSelectRevision())
    );

    connect(
        branchList, SIGNAL(activated(int)),
        this, SLOT(selectBranchOption())
    );

    connect(
        tagList, SIGNAL(activated(int)),
        this, SLOT(selectTagOption())
    );

    connect(
        selectedRevision, SIGNAL(textChanged(const QString &)),
        this, SLOT(selectSelectorOption())
    );

    connect(
        browseDirs, SIGNAL(clicked()),
        this, SLOT(openDirBrowser())
    );
}

CheckoutRevision::~CheckoutRevision()
{
    delete branchesModel;
    delete tagsModel;
    delete selectorCompleter;
}

void CheckoutRevision::accept()
{
    QString selector;

    if (checkoutByBranch->isChecked())
    {
        QString selection(branchList->currentText());
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No branch selected"),
                tr("Please select a branch from the list of branches."),
                QMessageBox::Ok
             );
             return;
        }
        selector = "h:" + selection;
    }

    if (checkoutByTag->isChecked())
    {
        QString selection(tagList->currentText());
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No tag selected"),
                tr("Please select a tag from the list of tags (if any are available)."),
                QMessageBox::Ok
             );
             return;
        }
        selector = "t:" + selection;
    }

    if (checkoutByRevision->isChecked())
    {
        QString selection = selectedRevision->text();
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No selector given"),
                tr("Please enter a monotone selector or choose a revision by clicking on 'select revision'."),
                QMessageBox::Ok
             );
             return;
        }
        selector = selection;
        Settings::addItemToList("RecentCheckoutSelectors", selection, 10);
    }

    QStringList revList(MonotoneUtil::resolveSelector(databaseFile, selector));

    if (revList.size() == 0)
    {
        QMessageBox::warning(
            this,
            tr("No revisions matched"),
            tr("Your selection didn't match any revisions or was invalid."),
            QMessageBox::Ok
         );
         return;
    }

    if (revList.size() > 1)
    {
        QMessageBox::information(
            this,
            tr("Multiple revisions matched"),
            tr("Your selection matched multiple revisions. Please choose one."),
            QMessageBox::Ok
        );

        emit selectRevisionFromSelector(selector);
        return;
    }

    QString checkoutPath = localCheckoutPath->text();
    QDir checkoutDir(checkoutPath);

    // check if the dir exists, and if not, create it
    if (!checkoutDir.exists())
    {
        QMessageBox::StandardButton btn = QMessageBox::question(
            this,
            tr("Directory does not exist"),
            tr("The checkout directory does not exist, "
               "do you want to create it?"),
            QMessageBox::Yes | QMessageBox::No
         );

         if (btn == QMessageBox::No) return;

         checkoutDir.setPath(QDir::rootPath());
         if (!checkoutDir.mkpath(checkoutPath))
         {
             QMessageBox::critical(
                this,
                tr("Directory path could not be created"),
                tr("The directory path could not be created, "
                   "please check your permissions or enter a different one."),
                QMessageBox::Ok
             );
             return;
         }
         checkoutDir.setPath(checkoutPath);
    }

    // check permissions
    QFileInfo info(checkoutDir.absolutePath());
    if (!info.isWritable())
    {
         QMessageBox::critical(
            this,
            tr("Directory not writable"),
            tr("The directory is not writable, "
               "please check your permissions or enter a different one."),
            QMessageBox::Ok
         );
         return;
    }

    QDir::Filters filters = QDir::NoDotAndDotDot | QDir::AllEntries |
                            QDir::Hidden | QDir::System;
    // ensure the user acknowledges that a directory may not be empty
    if (checkoutDir.entryList(filters).size() > 0)
    {
        QMessageBox::StandardButton btn = QMessageBox::question(
            this,
            tr("Directory not empty"),
            tr("The checkout directory is not empty. "
               "Do you want to continue?"),
            QMessageBox::Yes | QMessageBox::No
         );

         if (btn == QMessageBox::No) return;
    }

    WorkspaceCreator creator(databaseFile, revList.at(0), checkoutDir);

    if (!creator.run())
    {
        QMessageBox::critical(
            this,
            tr("Checkout failed"),
            tr("The checkout failed - please report this problem to "
               "the author and attach the guitone.log in your home directory."),
            QMessageBox::Ok
        );
        return;
    }

    emit revisionCheckedOut(checkoutDir.absolutePath());
    done(0);
}

void CheckoutRevision::slotSelectRevision()
{
    emit selectRevisionFromSelector(QString());
}

void CheckoutRevision::setSelectedRevision(const QString & revision)
{
    selectedRevision->setText(revision);
}

void CheckoutRevision::selectBranchOption()
{
    checkoutByBranch->setChecked(true);
}

void CheckoutRevision::selectTagOption()
{
    checkoutByTag->setChecked(true);
}

void CheckoutRevision::selectSelectorOption()
{
    checkoutByRevision->setChecked(true);
}

void CheckoutRevision::openDirBrowser()
{
    QString fn = QFileDialog::getExistingDirectory(
        this, tr("Select a directory for checkout...")
    );

    if (!fn.isEmpty())
    {
        localCheckoutPath->setText(fn);
    }
}

