/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "DatabaseDialogManager.h"
#include "vocab.h"

DatabaseDialogManager::DatabaseDialogManager(QWidget * parent)
  : DialogManager(parent), changesetBrowser(0),
    checkoutRevision(0), fileDiff(0), fileHistory(0), generateKeypair(0),
    keyManagement(0), revisionDiff(0), revisionManifest(0), selectRevision(0)
{}

DatabaseDialogManager::~DatabaseDialogManager()
{
    cleanup();
}

void DatabaseDialogManager::cleanup()
{
    if (changesetBrowser)   { delete changesetBrowser; changesetBrowser = 0; }
    if (checkoutRevision)   { delete checkoutRevision; checkoutRevision = 0; }
    if (fileDiff)           { delete fileDiff; fileDiff = 0; }
    if (fileHistory)        { delete fileHistory; fileHistory = 0; }
    if (generateKeypair)    { delete generateKeypair; generateKeypair = 0; }
    if (keyManagement)      { delete keyManagement; keyManagement = 0; }
    if (revisionDiff)       { delete revisionDiff; revisionDiff = 0; }
    if (revisionManifest)   { delete revisionManifest; revisionManifest = 0; }
    if (selectRevision)     { delete selectRevision; selectRevision = 0; }
}

void DatabaseDialogManager::closeAllDialogs()
{
    if (changesetBrowser)   changesetBrowser->close();
    if (checkoutRevision)   checkoutRevision->close();
    if (fileDiff)           fileDiff->close();
    if (fileHistory)        fileHistory->close();
    if (generateKeypair)    generateKeypair->close();
    if (keyManagement)      keyManagement->close();
    if (revisionDiff)       revisionDiff->close();
    if (revisionManifest)   revisionManifest->close();
    if (selectRevision)     selectRevision->close();

    DialogManager::closeAllDialogs();
}

void DatabaseDialogManager::init(const DatabaseFile & db)
{
    databaseFile = db;
    cleanup();
}

void DatabaseDialogManager::showChangesetBrowser()
{
    if (!changesetBrowser)
    {
        changesetBrowser = new ChangesetBrowser(parentWidget(), databaseFile);

        connect(
            changesetBrowser, SIGNAL(revisionManifest(const QString &)),
            this, SLOT(showRevisionManifest(const QString &))
        );

        connect(
            changesetBrowser, SIGNAL(diffFile(const QString &, const QString &, const QString &)),
            this,  SLOT(showFileDiff(const QString &, const QString &, const QString &))
        );
    }

    showDialog(changesetBrowser);
}

void DatabaseDialogManager::showCheckoutRevision()
{
    if (!checkoutRevision)
    {
        checkoutRevision    = new CheckoutRevision(parentWidget(), databaseFile);

        connect(
            checkoutRevision, SIGNAL(selectRevisionFromSelector(const QString &)),
            this, SLOT(showSelectRevision(const QString &))
        );

        // delegate this further
        connect(
            checkoutRevision, SIGNAL(revisionCheckedOut(const QString &)),
            this, SIGNAL(revisionCheckedOut(const QString &))
        );
    }

    showDialog(checkoutRevision);
}

void DatabaseDialogManager::showFileDiff(const QString & file, const QString & base, const QString & target)
{
    if (!fileDiff)
    {
        fileDiff = new FileDiff(parentWidget());
    }

    fileDiff->forDatabase(databaseFile, file, base, target);
    showDialog(fileDiff);
}

void DatabaseDialogManager::showFileHistory(const QString & file, const QString & startRevision)
{
    if (!fileHistory)
    {
        fileHistory = new FileHistory(parentWidget(), databaseFile);

        connect(
            fileHistory, SIGNAL(fileDiff(const QString &, const QString &, const QString &)),
            this, SLOT(showFileDiff(const QString &, const QString &, const QString &))
        );
    }

    fileHistory->readHistory(file, startRevision);
    showDialog(fileHistory);
}

void DatabaseDialogManager::showGenerateKeypair()
{
    if (!generateKeypair)
    {
        generateKeypair  = new GenerateKeypair(parentWidget(), databaseFile);

        if (keyManagement)
        {
            connect(
                generateKeypair, SIGNAL(keypairGenerated()),
                keyManagement, SLOT(readKeys())
            );
        }
    }

    showDialog(generateKeypair);
}

void DatabaseDialogManager::showKeyManagement()
{
    if (!keyManagement)
    {
        keyManagement = new KeyManagement(parentWidget(), databaseFile);

        connect(
            keyManagement, SIGNAL(generateKeypair()),
            this, SLOT(showGenerateKeypair())
        );
    }

    keyManagement->readKeys();
    showDialog(keyManagement);
}

void DatabaseDialogManager::showRevisionDiff(const QString & file, const QString & base, const QString & target)
{
    if (!revisionDiff)
    {
        revisionDiff = new RevisionDiff(parentWidget());

        connect(
            revisionDiff, SIGNAL(fileDiff(const QString &, const QString &, const QString &)),
            this, SLOT(showFileDiff(const QString &, const QString &, const QString &))
        );
    }

    revisionDiff->forDatabase(databaseFile, file, base, target);
    showDialog(revisionDiff);
}

void DatabaseDialogManager::showRevisionManifest(const QString & revision)
{
    if (!revisionManifest)
    {
        revisionManifest    = new RevisionManifest(parentWidget(), databaseFile);

        connect(
            revisionManifest, SIGNAL(fileHistory(const QString &, const QString &)),
            this, SLOT(showFileHistory(const QString &, const QString &))
        );
    }

    revisionManifest->readManifest(revision);
    showDialog(revisionManifest);
}

void DatabaseDialogManager::showSelectRevision(const QString & selector)
{
    if (!selectRevision)
    {
        selectRevision = new SelectRevision(parentWidget(), databaseFile);

        connect(
            selectRevision, SIGNAL(revisionDiff(const QString &, const QString &, const QString &)),
            this, SLOT(showRevisionDiff(const QString &, const QString &, const QString &))
        );

        connect(
            selectRevision, SIGNAL(revisionManifest(const QString &)),
            this, SLOT(showRevisionManifest(const QString &))
        );

        if (checkoutRevision)
        {
            connect(
                selectRevision, SIGNAL(revisionSelected(const QString &)),
                checkoutRevision, SLOT(setSelectedRevision(const QString &))
            );
        }
    }

    if (selector.size() > 0)
    {
        selectRevision->queryRevisions(selector);
    }

    showDialog(selectRevision);
}

