/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "FileDiff.h"
#include "MonotoneUtil.h"
#include "Settings.h"

#include <QHeaderView>
#include <QMessageBox>
#include <QScrollBar>
#include <QTextCodec>

FileDiff::FileDiff(QWidget * parent)
  : Dialog(parent), diffModel(0), fileModel(0), fileProxyModel(0)
{
    setupUi(this);
    Dialog::init();

    // setup correct header widths
    diffView->header()->setDefaultSectionSize(40);
    diffView->header()->setResizeMode(0, QHeaderView::Interactive);
    diffView->header()->setResizeMode(1, QHeaderView::Interactive);
    diffView->header()->setResizeMode(2, QHeaderView::ResizeToContents);

    QScrollBar * vertBar = diffView->verticalScrollBar();
    connect(
        vertBar, SIGNAL(valueChanged(int)),
        this, SLOT(changeCurrentIndex(int))
    );

    connect(
        firstRevision, SIGNAL(toggled(bool)),
        this, SLOT(versionToggled())
    );
    connect(
        secondRevision, SIGNAL(toggled(bool)),
        this, SLOT(versionToggled())
    );
    connect(
        bothRevisions, SIGNAL(toggled(bool)),
        this, SLOT(versionToggled())
    );

    connect(
        scrollToNext, SIGNAL(clicked()),
        this, SLOT(getNextGroup())
    );

    connect(
        scrollToPrev, SIGNAL(clicked()),
        this, SLOT(getPrevGroup())
    );

    fileModel = new GetFile(this);
    fileProxyModel = new GetFileProxyModel(this);
    fileProxyModel->setSourceModel(fileModel);

    diffModel = new ContentDiff(this);

    diffView->setModel(fileProxyModel);
    diffStatusView->setModel(fileProxyModel);

    connect(
        fileModel, SIGNAL(fileRead()),
        this, SLOT(applyDiff())
    );

    connect(
        diffModel, SIGNAL(diffRead()),
        this, SLOT(applyDiff())
    );

    ByteArrayList codecNamesAndAliases = QTextCodec::availableCodecs();
    QSet<QByteArray> codecNameSet;

    // eliminate aliases
    foreach (QByteArray codec, codecNamesAndAliases)
    {
        codecNameSet.insert(QTextCodec::codecForName(codec)->name());
    }

    ByteArrayList codecNames = codecNameSet.toList();
    qSort(codecNames);

    foreach (QByteArray codec, codecNames)
    {
        //! FIXME: better ordering, i.e. popular ones on top
        textEncoding->addItem(QString(codec));
    }

    connect(
        textEncoding, SIGNAL(currentIndexChanged(const QString &)),
        this, SLOT(switchOutputEncoding(const QString &))
    );
}

void FileDiff::forDatabase(const DatabaseFile & db, const QString & file,
                           const QString & base, const QString & target)
{
    fileName = file;
    loaded = false;
    workspace = QString();

    QString encoding = "UTF-8";
    if (!MonotoneUtil::getAttribute(db, base, file, "guitone:file-encoding",
                                    encoding))
    {
        L(QString("could not query file encoding of %1, using default").arg(file));
    }
    textEncoding->setCurrentIndex(textEncoding->findText(encoding));

    setWindowTitle(tr("File differences of \"%1\"").arg(fileName));

    firstRevision->setText(
        tr("only left (%1)").arg(base.left(12).append("..."))
    );

    secondRevision->setText(
        tr("only right (%1)").arg(target.left(12).append("..."))
    );

    fileModel->readFileByName(db, fileName, base, encoding);
    diffModel->readDatabaseDiff(db, fileName, base, target, encoding);
}

void FileDiff::forWorkspace(const WorkspacePath & ws, const QString & file,
                            const QString & base, const QString & target)
{
    fileName = file;
    loaded = false;
    workspace = ws;

    QString encoding = "UTF-8";
    QPair<QString, QString> attrval;
    if (!MonotoneUtil::getAttribute(ws, file, "guitone:file-encoding", attrval) ||
        attrval.second == "dropped")
    {
        L(QString("could not query file encoding of %1, using default").arg(file));
    }
    else
    {
        encoding = attrval.first;
    }
    textEncoding->setCurrentIndex(textEncoding->findText(encoding));

    setWindowTitle(tr("File differences of \"%1\"").arg(fileName));

    QString left = tr("workspace parent");
    QString right = tr("workspace revision");

    if (!base.isEmpty())
    {
        left = base.left(12).append("...");
    }

    if (!target.isEmpty())
    {
        right = target.left(12).append("...");
    }

    firstRevision->setText(tr("only left (%1)").arg(left));
    secondRevision->setText(tr("only right (%1)").arg(right));

    DatabaseFile db = MonotoneUtil::getDatabaseFile(ws);
    if (base.isEmpty())
    {
        QString workspaceParent = MonotoneUtil::getBaseWorkspaceRevision(ws);
        fileModel->readFileByName(db, fileName, workspaceParent, encoding);
        diffModel->readWorkspaceDiff(ws, fileName, workspaceParent, target, encoding);
        return;
    }

    fileModel->readFileByName(db, fileName, base, encoding);
    diffModel->readWorkspaceDiff(ws, fileName, base, target, encoding);
}

FileDiff::~FileDiff()
{
    if (fileModel)      delete fileModel;
    if (fileProxyModel) delete fileProxyModel;
    if (diffModel)      delete diffModel;
}

void FileDiff::applyDiff()
{
    if (!loaded)
    {
        loaded = true;
        return;
    }

    Diff * diff = diffModel->getDiff(fileName);
    if (diff == 0)
    {
        QMessageBox::information(
            this,
            tr("File has not changed"),
            tr("The file has not been changed between these revisions."),
            QMessageBox::Ok
        );
        reject();
        return;
    }

    if (diff->is_binary)
    {
        QMessageBox::information(
            this,
            tr("File is binary"),
            tr("The file is binary and cannot be diffed."),
            QMessageBox::Ok
        );
        reject();
        return;
    }

    fileModel->applyDiff(diff);
    diffStatusView->update();
}

void FileDiff::versionToggled()
{
    if (firstRevision->isChecked())
    {
        fileProxyModel->setFileVersion(GetFileProxyModel::Left);
        diffView->header()->showSection(0);
        diffView->header()->hideSection(1);
    }
    else if (secondRevision->isChecked())
    {
        fileProxyModel->setFileVersion(GetFileProxyModel::Right);
        diffView->header()->showSection(1);
        diffView->header()->hideSection(0);
    }
    else if (bothRevisions->isChecked())
    {
        fileProxyModel->setFileVersion(GetFileProxyModel::Both);
        diffView->header()->showSection(0);
        diffView->header()->showSection(1);
    }

    diffStatusView->update();
}

void FileDiff::scrollToGroup(bool forward)
{
    if (!currentIndex.isValid()) return;
    QModelIndex proxyIndex = fileProxyModel->mapFromSource(currentIndex);

    // recursion: if the queried group is not valid in the view because it
    // has been filtered out, get the next group and try again
    if (!proxyIndex.isValid())
    {
        if (forward)
        {
            getNextGroup();
            return;
        }
        getPrevGroup();
        return;
    }

    diffView->scrollTo(proxyIndex, QAbstractItemView::PositionAtTop);
}

void FileDiff::getPrevGroup()
{
    currentIndex = fileModel->getPrevGroup(currentIndex);
    scrollToGroup(false);
}

void FileDiff::getNextGroup()
{
    currentIndex = fileModel->getNextGroup(currentIndex);
    scrollToGroup(true);
}

void FileDiff::changeCurrentIndex(int pos)
{
    if (!fileProxyModel) return;

    QModelIndex newCurrent = fileProxyModel->mapToSource(
         fileProxyModel->index(pos, 0, QModelIndex())
    );

    if (newCurrent.isValid())
    {
        currentIndex = newCurrent;
    }
}

void FileDiff::switchOutputEncoding(const QString & encoding)
{
    // if we're not yet fully loaded the models, don't do anything
    if (!loaded) return;

    fileModel->switchOutputEncoding(encoding);
    diffModel->switchOutputEncoding(encoding);

    // the model data are converted instantly, so we don't have
    // to wait for signals this time...
    applyDiff();

    if (!workspace.isEmpty())
    {
        MonotoneTask in(QStringList() << "set_attribute"
                                      << fileName
                                      << "guitone:file-encoding"
                                      << encoding);
        MonotoneTask out = MonotoneUtil::runSynchronousWorkspaceTask(workspace, in);

        if (out.getReturnCode() != 0)
        {
            W(QString("couldn't set guitone:file-encoding on %1: %2")
                .arg(fileName).arg(out.getDecodedOutput()));
        }
    }
}

